/* OLED SSD 1306 (128x64) IIC-Interface library 

License: GPL V3
written by Sven K. (oog / proog.de)

History:
YYYY-MM-DD
2018-10:
  first lib-version
2018-10-19:
  Print text within screen range
  now scrolls without buffer

*/


#include "oled_ssd1306.h"
#include <Wire.h> // I2C lib

//OLED::OLED() {
//}

void OLED::D1306_cmd(uint8_t c) {
  // I2C
  uint8_t control = 0x00;   // Co = 0, D/C = 0
  Wire.beginTransmission(_i2caddr);
  Wire.write(control);
  Wire.write(c);
  uint8_t r = Wire.endTransmission();
  if (r) {
    if (oledOk) {
      Serial.print(F(" IIC Error Addr=0x"));
      Serial.print(_i2caddr, 16);
      Serial.print(F(", Status="));
      Serial.println(r);
      oledOk = 0;
    }
  }
}

void OLED::D1306_cmd(uint8_t c, uint8_t d) {
  // I2C command with parameter in one I2C telegram
  uint8_t control = 0x00;   // Co = 0, D/C = 0
  Wire.beginTransmission(_i2caddr);
  Wire.write(control);
  Wire.write(c);
  Wire.write(d);
  uint8_t r = Wire.endTransmission();
  if (r) {
    if (oledOk) {
      Serial.print(F(" IIC Error Addr=0x"));
      Serial.print(_i2caddr, 16);
      Serial.print(F(", Status="));
      Serial.println(r);
      oledOk = 0;
    }
  }
}


// initialize OLED display

void OLED::D1306_begin(uint8_t i2caddr) {
  _i2caddr = i2caddr;

  //  Software Initialization Flow Chart

  D1306_cmd(D1306_DISPLAYON | 0);

  //  Enable charge pump regulator (8Dh, 14h)
  D1306_cmd(D1306_CHARGEPUMP, 0x14);

  //  Set MUX Ratio (A8h, 3Fh)
  D1306_cmd(D1306_MUXRATIO, D1306_HEIGHT - 1);

  //  Set Display Offset (D3h, 00h)
  D1306_cmd(D1306_DISPLAYOFFSET, 0x0);

  //  Set Display Start Line (40h)
  D1306_cmd(D1306_STARTLINE | 0x0);

  //  Set Segment re-map (A0h/A1h)
  D1306_cmd(D1306_SEGREMAP | 0x1);
  //D1306_cmd(D1306_SEGREMAP | 0x0);

  //  Set COM Output Scan Direction (C0h/C8h)
  //D1306_cmd(D1306_COMSCAN);
  D1306_cmd(D1306_COMSCANVERTFLIP);

  //  Set Contrast Control (81h, 7Fh)
  D1306_cmd(D1306_CONTRAST, 0x7F);

  //  Disable Entire Display On (A4h)
  D1306_cmd(D1306_ENTIREDISPON | 0);

  //  Set Normal Display (A6h)
  D1306_cmd(D1306_INVERSE | 0);

  //  Set Osc Frequency (D5h, 80h)
  D1306_cmd(D1306_DISPLAYCLKDIV, 0x80);

  //  Set COM Pins hardware configuration (DAh, 02)
  D1306_cmd(D1306_COMPINSHWCONFIG, 0x12); // 0x02

  //  Extra...
  D1306_cmd(D1306_MEMORYMODE, 0x00);
  //  D1306_cmd(D1306_DISPLAYCLKDIV, 0x80);
  //  D1306_cmd(D1306_PRECHARGEPERIOD, 0xF1);
  //  D1306_cmd(D1306_VCOMHADJUST, 0x40);

  //  Display On (AFh)
  D1306_cmd(D1306_DISPLAYON | 1);
}


int OLED::strPix(char s[]) {
  return (strlen(s)*scaleX*fontCharacterBytes);
}


void OLED::displayCursXY(uint8_t x, uint8_t y) {
  _displayCursX = x;
  _displayCursY = y;
  D1306_cmd(D1306_PGSTARTADDR | (y & 0x07));
  D1306_cmd(D1306_SETLOWCOLUMN | (x & 0x0f));
  D1306_cmd(D1306_SETHIGHCOLUMN | ((x >> 4) & 0x0f));
}


void OLED::displayFill(uint8_t d) {
  displayCursXY(0, 0);
  for (uint8_t i = 0; i < (D1306_WIDTH / 16 * D1306_HEIGHT / 8); i++) {
    Wire.beginTransmission(_i2caddr);
    Wire.write(0x40);
    for (uint8_t x = 0; x < 16; x++) {
      Wire.write(d);
    }
    Wire.endTransmission();
  }
}


void OLED::displayFill(uint8_t d1, uint8_t d2) {
  displayCursXY(0, 0);
  for (uint8_t i = 0; i < (D1306_WIDTH / 16 * D1306_HEIGHT / 8); i++) {
    Wire.beginTransmission(_i2caddr);
    Wire.write(0x40);
    for (uint8_t x = 0; x < 8; x++) {
      Wire.write(d1);
      Wire.write(d2);
    }
    Wire.endTransmission();
  }
}


void OLED::displayByte(uint8_t d) {
  Wire.beginTransmission(_i2caddr);
  Wire.write(0x40);
  Wire.write(d);
  Wire.endTransmission();
}


void OLED::displayContrast(uint8_t d) {
  D1306_cmd(D1306_CONTRAST, d);
}


/////////////////////////////////////////////////////
//
// Direct print to Display
//
/////////////////////////////////////////////////////


void OLED::displayBuf(void) {
  Wire.beginTransmission(_i2caddr);
  Wire.write(0x40);
  for (uint16_t i = 1; i <= _buf16[0]; i++) {
    Wire.write(_buf16[i]);
  }
  Wire.endTransmission();
}


void OLED::displayBufN(uint8_t ptr[], uint8_t count) {
  Wire.beginTransmission(_i2caddr);
  Wire.write(0x40);
  for (uint16_t i = 0; i < count; i++) {
    Wire.write(ptr[i]);
  }
  Wire.endTransmission();
}


// display (part of) character at position pos
// from ROM
// startX = first pixel line on screen or char start
// endX = last pixel line on screen or char end
// fontCharacterBytes * scaleX = char width

void OLED::displayRomChar(int startX, int endX, uint8_t c) {
  uint8_t d;
  uint8_t charWidth = fontCharacterBytes*scaleX;
  const uint8_t* p = font + (c - 32) * fontCharacterBytes;
  
  // up to 16 bytes in a IIC tranmission
  Wire.beginTransmission(_i2caddr);
  Wire.write(0x40);
  for (uint8_t i = startX; i < endX; i++) {
    d = charMask ^ pgm_read_byte_near(p+(i/scaleX));
    switch (scaleY) {
      
      // Y = 1 line
      case 1:
        Wire.write(d);
      break;
      
      // Y = 2 lines
      case 2:
        switch (charLineY) {
          case 0:
            Wire.write(stretch4bits[15 & d]);
          break;
          case 1:
            Wire.write(stretch4bits[d >> 4]);
          break;
        }
      break;
      
      // Y = 4 lines
      case 4:
        switch (charLineY) {
          case 0:
            Wire.write(stretch2bits[d & 3]);
          break;
          case 1:
            Wire.write(stretch2bits[(d >> 2) & 3]);
          break;
          case 2:
            Wire.write(stretch2bits[(d >> 4) & 3]);
          break;
          case 3:
            Wire.write(stretch2bits[d >> 6]);
          break;
        }
      break;
    }
  }
  Wire.endTransmission();
}


// display font size and format
// fntsize:
//   --6x8----8x8-- font
// 1:  6x 8 ( 8x 8)
// 2:  6x16 ( 8x16)
// 3: 12x16 (16x16)
// 4: 12x32 (16x32)
// 5: 24x32 (32x32)
//
// format:
// NORMAL
// INVERSE
// UNDERLINE
// INVERSE+UNDERLINE


void OLED::displayFormat(uint8_t fntsize, uint8_t format) {
//  _charFontsize = fntsize;
  if (format & INVERSE) {
    charMask = 255;
  } else {
    charMask = 0;
  }
  if (format & UNDERLINE) {
    charMask ^= 128;
  }
  // fntsize = 1 .. 5
  charLineY = 0;
  switch (fntsize) {
    case 1:
      scaleX = 1;
      scaleY = 1;
    break;
    case 2:
      scaleX = 1;
      scaleY = 2;
    break;
    case 3:
      scaleX = 2;
      scaleY = 2;
    break;
    case 4:
      scaleX = 2;
      scaleY = 4;
    break;
    case 5:
      scaleX = 4;
      scaleY = 4;
    break;
  }
}


void OLED::displayScale(uint8_t scale_X, uint8_t scale_Y) {
  scaleX = scale_X;
  scaleY = scale_Y;
}


void OLED::dispOneLine(int px,int py,int xs,int xe,int firstchar,int lastchar,char s[]) {
  int i = firstchar;
  int charW = (scaleX*fontCharacterBytes);
  if (py >= (D1306_HEIGHT/8)) return;
  if (py < 0) return;
  displayCursXY(px, py);
  
  // first character
  if ((xs!=0) && (s[i] != 0))
    displayRomChar(xs,charW,s[i++]);
    
  // all characters in between
  while ((s[i] != 0) && (i < lastchar)) {
    displayRomChar(0,charW,s[i++]);
  }
  
  // last character
  if ((xe!=0) && (s[i] != 0))
    displayRomChar(0,xe,s[i++]);
}


void OLED::displayPrintXY(int x, int y, char s[]) {
  int xs = 0, xe = 0;
  int charW = (scaleX*fontCharacterBytes);
  int px = x;
  int slen = strlen(s);
  int firstchar = 0;
  int lastchar = 0;
  
  // Check coordinates
  // when out of screen, then return
  if (x >= D1306_WIDTH) return;
  if ((x + slen*charW) <= 0) return;
  
  // find first character that prints on screen
  if (x<0) {
    firstchar = -x / charW;
    px = x + firstchar*charW;
    if (px<0) {
      xs = -px;
      px = 0;
    }
  }
  
  // find last character that prints on screen
  lastchar = (D1306_WIDTH-x)/charW;
  if (lastchar<=strlen(s)) {
    //xe = (D1306_WIDTH+charW-x) % charW;
    xe = (D1306_WIDTH+(slen*charW)-x) % charW;
  }
  
  // scaleY = 1, 2, 4
  switch (scaleY) {
    case 1:
      charLineY = 0;
      dispOneLine(px,y,xs,xe,firstchar,lastchar,s);
    break;
    
    case 2:
      charLineY = 1;
      dispOneLine(px,y+1,xs,xe,firstchar,lastchar,s);
      charLineY = 0;
      dispOneLine(px,y,xs,xe,firstchar,lastchar,s);
    break;
    
    case 4:
      charLineY = 3;
      dispOneLine(px,y+3,xs,xe,firstchar,lastchar,s);
      charLineY = 2;
      dispOneLine(px,y+2,xs,xe,firstchar,lastchar,s);
      charLineY = 1;
      dispOneLine(px,y+1,xs,xe,firstchar,lastchar,s);
      charLineY = 0;
      dispOneLine(px,y,xs,xe,firstchar,lastchar,s);
    break;
  }
}


void OLED::displayPrintXY(int x, int y, const __FlashStringHelper *s) {
//  const int _charperline = D1306_WIDTH / fontCharacterBytes;
  // max constant string length = 80 characters
  const int _charperline = 80;
  char st[_charperline + 1];
  int i = 0;

  PGM_P p = reinterpret_cast<PGM_P>(s);
  while (i < _charperline) {
    st[i] = pgm_read_byte(p++);
    if (st[i] == 0) break;
    i++;
  }
  st[i] = 0;
  displayPrintXY(x, y, &st[0]);
}


void OLED::displayXYUInt(uint8_t x, uint8_t y, uint16_t i, uint8_t digits) {
  char digitstr[6] = "00000";
  if (digits < 1) return;
  if (digits > 5) return;
  digitstr[4] = '0' + ((i / 1) % 10);
  digitstr[3] = '0' + ((i / 10) % 10);
  digitstr[2] = '0' + ((i / 100) % 10);
  digitstr[1] = '0' + ((i / 1000) % 10);
  digitstr[0] = '0' + ((i / 10000) % 10);
  displayPrintXY(x, y, &digitstr[5] - digits);
}


void OLED::displayClrLine(uint8_t y) {
  for (uint8_t l = 0; l < scaleY; l++) {
    displayCursXY(0, y + l);
    for (uint8_t i = 0; i < (D1306_WIDTH / 16); i++) {
      Wire.beginTransmission(_i2caddr);
      Wire.write(0x40);
      for (uint8_t x = 0; x < 16; x++) {
        Wire.write(0);
      }
      Wire.endTransmission();
    }
  }
}


void OLED::scrollText(char* s, uint8_t y) {
  for (int i = D1306_WIDTH; i > -strPix(s); i=i-2) {
    displayPrintXY(i,y,s);
    delay(1);
  }
}


void OLED::scrollTextFast(char* s, uint8_t y) {
  for (int i = D1306_WIDTH; i > -strPix(s); i=i-3) {
    displayPrintXY(i,y,s);
    if((i&1) == 0) delay(1);
  }
}

